/*
 * Copyright (c) 1990 Regents of Bradley University.
 * All rights reserved.  
 * Author : Krishnakumar Siva
 */

/*
 * printjob -- print jobs in the queue.
 *
 */
#include "lp.h"
#include <malloc.h>

/*
 * Error tokens
 */
#define REPRINT		-2
#define ERROR		-1
#define	OK		0
#define	FATALERR	1
#define	NOACCT		2
#define	FILTERERR	3    	/* can't handle filter's in this version */
#define	ACCESS		4
#define WAIT		5
#define QUEUE		6
#define SPACE		7
#define FULL		8
#define COPIES		9

char	title[80];		/* ``pr'' title */
int	pfd;			/* printer file descriptor */
int	ofd;			/* output filter file descriptor */
int	lfd;			/* lock file descriptor */
int	pid;			/* pid of lpd process */
int	prchild;		/* id of pr process */
int	child;			/* id of any filters */
int	ofilter;		/* id of output filter, if any */
int	tof;			/* true if at top of form */
int	remote;			/* true if sending files to remote */
dev_t	fdev;			/* device of file pointed to by symlink */
ino_t	fino;			/* inode of file pointed to by symlink */

short 	RW;
char 	*ST;
char	line[BUFSIZ];
char	fromhost[32];		/* user's host machine */
char	logname[32];		/* user's login name */
extern  char *jobname;		 /* job or file name */
extern  char *user;
char	class[32];		/* classification field */
char	width[10] = "-w";	/* page width in characters */
char	length[10] = "-l";	/* page length in lines */
char	pxwidth[10] = "-x";	/* page width in pixels */
char	pxlength[10] = "-y";	/* page length in pixels */
char	indent[10] = "-i0";	/* indentation size in characters */
char	tempfile[] = "errsXXXXXX"; /* file name for filter output */
struct passwd 	*getpwnam();
extern  FILE    *fd;
extern  int	num_cmds;
extern  char    commands[4096];
extern  char	*cfname;
extern  char	*tfname;
extern  char 	*dfname; 
extern  char 	*arg;
extern 	void	cleanup();
/*
 * Send the daemon control file (cf) and any data files.
 * If any error encountered send a mail message by creating a pipe
 * and print the nature of the message to the piped process
 */
sendit()
{
	register int i, err = OK;
	char *cp, last[BUFSIZ];
	int linel  = 0;
	int lineno = 0;
	int c = 0;
	FILE *cfp;

	/*
	 *      read the control file for work to do
	 *
	 *      file format -- first character in the line is a command
	 *      rest of the line is the argument.
	 *      commands of interest are:
	 *
	 *            a-z -- "file name" name of file to print
	 */
			switch (sendfile('\3', arg)) {
			case OK:
				break;
			case REPRINT:
				sendmail(user, REPRINT);
				cleanup("");            	
			case ACCESS:
				sendmail(user, ACCESS);
				cleanup("");            	
			case ERROR:
				sendmail(user, ERROR);
				err = ERROR;
				cleanup("");            	
			case SPACE:
				sendmail(user,SPACE);
				cleanup("");
			case FULL:
				 sendmail(user,FULL);
				cleanup("");
			}
	if (err == OK && sendfile('\2', tfname) > 0) {
		(void) fclose(fd);
		sendmail(user, ACCESS);
		cleanup("");            	
		return(REPRINT);
	}
	return(err);
}

/*
 * Send a data file to the remote machine and spool it.
 * Return positive if we should try resending.
 */
sendfile(type, file)
	char type, *file;
{
	register int f, i,j, amt, writvar;
	struct stat stb;
	char buf[BUFSIZ];
	int sizerr, resp;

#ifdef SYSTEM5
	if ((stat(file, &stb) < 0 || (f = open(file, O_RDONLY)) < 0) && type == '\3') 
#else
	if (lstat(file, &stb) < 0 || (f = open(file, O_RDONLY)) < 0)
#endif
		return(ERROR);
	/*
	 * Check to see if data file is a symbolic link. If so, it should
	 * still point to the same file or someone is trying to print something
	 * he shouldn't.
	 */
#ifndef SYSTEM5
	if ((stb.st_mode & S_IFMT) == S_IFLNK && fstat(f, &stb) == 0 &&
	    (stb.st_dev != fdev || stb.st_ino != fino) && type == '\3')
		return(ACCESS);
#endif
	if (type == '\3') {
		file = dfname; }
	if (type == '\2') {
		file = cfname;
		stb.st_size = strlen(commands);
	}
 	(void) sprintf(buf, "%c%d %s\n", type, stb.st_size, file);
	amt = strlen(buf);
	for (i = 0;  ; i++) {
		writvar = write(pfd, buf, amt);
		resp = response();
		if ( writvar != amt || resp < 0 || resp == '\1') {
			(void) close(f);
			return(REPRINT);
		} else if (resp == '\0')
			break;
		if (i == 0)
			return SPACE;	
		if (i == 10)
			return FULL;	
		sleep(5 * 60);
	}
	if (i)
		status("sending to %s", RM);
	sizerr = 0;
	for (i = 0; i < stb.st_size && type == '\3'; i += BUFSIZ) {
		amt = BUFSIZ;
		if (i + amt > stb.st_size)
			amt = stb.st_size - i;
		if (sizerr == 0 && read(f, buf, amt) != amt)
			sizerr = 1;
		if (write(pfd,buf,amt) != amt) {
			(void) close(f);
			return(REPRINT);
		}
	}
	amt = 255;
	if (type == '\2')  
		for (i = 0; i < strlen(commands); i += amt) {
			if (i + amt > stb.st_size)
				amt = stb.st_size - i;
			for (j = 0; j < amt; j++)
			   buf[j] = commands[i+j];
			if (write(pfd, buf, amt) != amt) {
				return(REPRINT); }
		}
	else
		(void) close(f);
	if (sizerr) {
		fprintf(stderr,"%s", "Error in sending\n");
		/* tell recv job to ignore this file */
		(void) write(pfd, "\1", 1);
		return(ERROR);
	}
	if (write(pfd, "", 1) != 1 || response()) {
		return(REPRINT); }
	return(OK);
}

/*
 * Check to make sure there have been no errors and that both programs
 * are in sync with eachother.
 * Return non-zero if the connection was lost.
 */
response()
{
	char resp;
        int tmpvar;

	tmpvar = read(pfd, &resp, 1);
	if ( tmpvar < 0 ) {
		fprintf(stderr,"Lost connection for %s\n",printer);
		return(-1);
	}
	return(resp);
}




/*
 * sendmail ---
 *   tell people about job completion
 */
sendmail(user, bombed)
	char *user;
	int bombed;
{
	register int i;
	int p[2], s;
	register char *cp;
	char buf[100];
	struct stat stb;
	FILE *fp;

	sprintf(buf,"/usr/lib/sendmail -f\"Line Printer Manager\" %s",user);
	setuid(0);
	if ((fp = popen(buf,"w")) == NULL) 
		perror("Piping failed");
		fprintf(fp,"To: %s\n", user);
		fprintf(fp,"Subject: printer job\n\n");
		fprintf(fp,"Your printer job ");
		if (*jobname)
			fprintf(fp,"(%s) ", jobname);
		switch (bombed) {
		case COPIES:
			fprintf(fp,"\nMaximum number of copies for a job is 256\n");
			break;
		case OK:
			fprintf(fp,"\nCompleted successfully\n");
			break;
		default:
		case FATALERR:
			fprintf(fp,"\ncould not be printed\n");
			break;
		case NOACCT:
			fprintf(fp,"\ncould not be printed without an account on %s\n", host);
			break;
		case FILTERERR:
/*
			if (stat(tempfile, &stb) < 0 || stb.st_size == 0 ||
			    (fp = fopen(tempfile, "r")) == NULL) {
				fprintf(fp,"\nwas printed but had some errors\n");
				break;
			}
			fprintf(fp,"\nwas printed but had the following errors:\n");
			while ((i = getc(fp)) != EOF)
				putchar(i);
			(void) fclose(fp);
*/
			break;
		case ACCESS:
			fprintf(fp,"\nWas not printed because it was not linked to the original file\n");
			break;
		case REPRINT:
			fprintf(fp,"\nCould not be printed. Reprint the file.\n");
			break;
		case ERROR:
			fprintf(fp,"\nWas not printed due to remote problem\n");
			break;
		case WAIT:
			fprintf(fp,"\nWaiting for remote printer \"%s\" at \"%s\" to accept\n",printer,RM);
			fprintf(fp,"\nCause of the problem: Printer down or queuing disabled\n");
			break;
		case QUEUE:
			fprintf(fp,"\nWaiting for queue to be enabled on %s\n",RM);
			break;
		
		case SPACE: 
			fprintf(fp,"No space on %s; waiting for queue to drain\n",RM); 
			break;
		case FULL:
			fprintf(fp,"%s: can't send to %s; queue full",printer,RM);
	}
	fclose(fp);
	cleanup("");
}



/*
 * Acquire line printer or remote connection.
 */
openpr(s)
int s;
{
	register int i, n;
	int resp;
	
	if (*LP) {
		for (i = 1; ; i = i < 32 ? i << 1 : i) {
			pfd = open(LP, RW ? O_RDWR : O_WRONLY);
			if (pfd >= 0)
				break;
			if (errno == ENOENT) {
				exit(1);
			}
			if (i == 1)
				status("waiting for %s to become ready (offline ?)", printer);
			sleep(i);
		}
/*
		if (isatty(pfd))
			setty();
		status("%s is ready and printing", printer);
*/
	} else if (RM != NULL) {
		for (i = 1; ; i = i < 256 ? i << 1 : i) {
			resp = -1;
			pfd = s;
			if (pfd >= 0) {
				(void) sprintf(line, "\2%s\n", RP);
			        n = strlen(line);
				if (write(pfd, line, n) == n  &&
                                    (resp = response()) == 0 ) 
					break; 
				(void) close(pfd);
			}
			if (i == 1) {
				if (resp == '\1')   {
					sendmail(user,WAIT);
					cleanup();
				}
				else {
					sendmail(user,QUEUE);
 					i = 256;   
 
				}
			}
/*
			sleep(i);
*/
		}
		status("sending to %s", RM);
		remote = 1;
	} else {
		fprintf(stderr, "%s: no line printer device or host name",printer);
		exit(1);
	}
}

struct bauds {
	int	baud;
	int	speed;
} bauds[] = {
	50,	B50,
	75,	B75,
	110,	B110,
	134,	B134,
	150,	B150,
	200,	B200,
	300,	B300,
	600,	B600,
	1200,	B1200,
	1800,	B1800,
	2400,	B2400,
	4800,	B4800,
	9600,	B9600,
	19200,	EXTA,
	38400,	EXTB,
	0,	0
};


/*VARARGS1*/
status(msg, a1, a2, a3)
	char *msg;
{
/* CHANGE [ebd 20-may-87] use fcntl struct flock for system V */
#ifdef SYSTEM5
	struct flock flock;
#endif
	register int fd;
	char buf[BUFSIZ];

	umask(0);
	fd = open(ST, O_WRONLY|O_CREAT, 0664);
/* CHANGE [ebd 20-may-87] use fcntl for system V */
#ifdef SYSTEM5
	flock.l_type = F_WRLCK;
	flock.l_whence = 0;
	flock.l_start = 0;
	flock.l_len = 0;
	if (fd < 0 || fcntl(fd, F_SETLKW, &flock) < 0) {
/*
		syslog(LOG_ERR, "%s: %s: %m", printer, ST);
		exit(1);
*/
	}
#else
	if (fd < 0 || flock(fd, LOCK_EX) < 0) {
/*
		syslog(LOG_ERR, "%s: %s: %m", printer, ST);
		exit(1);
*/
	}
	ftruncate(fd, 0);
#endif
	sprintf(buf, msg, a1, a2, a3);
	strcat(buf, "\n");
	(void) write(fd, buf, strlen(buf));
	(void) close(fd);
}
